#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <stdexcept>

struct pixel
{
	pixel(int rr, int gg, int bb)
	{
		r = rr;
		g = gg;
		b = bb;
	}
	
	bool operator != (const pixel& rh) const
	{
		return (r != rh.r || g != rh.g || b != rh.b);
	}

	int r;
	int g;
	int b;
};

std::ostream& operator << (std::ostream& os, const pixel& p)
{
	os << p.r << ' ' << p.g << ' ' << p.b;
	return os;
}

int iabs(int x)
{
	return x >= 0 ? x : -x;
}

class diff_viewer
{
public:
	virtual pixel diff(const pixel&, const pixel&) const = 0;
	virtual ~diff_viewer() {}
};

class normal_viewer : public diff_viewer
{
public:
	pixel diff(const pixel& p1, const pixel& p2) const
	{
		return pixel (iabs (p1.r - p2.r), iabs (p1.g - p2.g), iabs (p1.b - p2.b));
	}
};

class enhance_viewer : public diff_viewer
{
	static const int ENH = 10;
public:
	pixel diff(const pixel& p1, const pixel& p2) const
	{
		return pixel (ENH * iabs (p1.r - p2.r),
			ENH * iabs (p1.g - p2.g),
			ENH * iabs (p1.b - p2.b));
	}
};

class classify_viewer : public diff_viewer
{
	int sqr(int x) const
	{
		return x*x;
	}
public:
	pixel diff(const pixel& p1, const pixel& p2) const
	{
		int sum = sqr (p1.r - p2.r)
			+ sqr (p1.g - p2.g)
			+ sqr (p1.b - p2.b);

		if (sum == 0) {
			return pixel (0, 0, 0);
		}
		else if (sum == 1) {
			return pixel (0, 0, 255);
		}
		else if (sum <= 3) {
			return pixel (0, 255, 0);
		}
		else if (sum <= 6) {
			return pixel (255, 255, 0);
		}
		else {
			return pixel (255, 0, 0);
		}
	}
};

class verify_viewer : public diff_viewer
{
	int sqr(int x) const
	{
		return x*x;
	}
public:
	pixel diff(const pixel& p1, const pixel& p2) const
	{
		int sum = sqr (p1.r - p2.r)
			+ sqr (p1.g - p2.g)
			+ sqr (p1.b - p2.b);

		if (sum <= 3) {
			return pixel (0, 0, 0);
		}
		else if (sum <= 6) {
			return pixel (255, 255, 0);
		}
		else {
			return pixel (255, 0, 0);
		}
	}
};

struct ppm
{
	ppm(int w, int h, const std::vector<pixel>& pv) :
		width(w), height(h), pixels(pv)
	{}

	const int width;
	const int height;
	const std::vector<pixel> pixels;
};

void usage()
{
	std::cerr << "usage: ppmdiff [option] file1 file2" << std::endl;
	std::cerr << "options:" << std::endl
		<< "  -e: enhance differences" << std::endl
		<< "  -c: classify differences" << std::endl
		<< "  -v: verify differences" << std::endl;
}

ppm read_ppm(const char* filename)
{
	std::string header;
	int w, h, d;
	
	std::ifstream ifs(filename);
	if (ifs.fail()) {
		throw std::runtime_error ("file not found");
	}
	
	ifs >> header;
	if (header != "P3") throw std::runtime_error ("ppm header error");
	
	ifs >> w >> h >> d;
	int r, g, b;
	std::vector<pixel> pv;
	
	while (ifs >> r >> g >> b) {
		pv.push_back(pixel(r, g, b));
		if (pv.size() >= w*h) break;
	}
	if (pv.size() < w*h) throw std::runtime_error ("missing pixels");
	
	return ppm(w, h, pv);
}

int max_error (const pixel& p0, const pixel& p1)
{
	return std::max(std::max(iabs(p0.r-p1.r), iabs(p0.g-p1.g)), iabs(p0.b-p1.b));
}

void compare (const diff_viewer* viewer, const ppm& ppm0, const ppm& ppm1)
{
	if (ppm0.width != ppm1.width || ppm0.height != ppm1.height) {
		std::cerr << "error: different image size" << std::endl;
		return;
	}
	
	std::cout << "P3" << std::endl;
	std::cout << ppm0.width << ' ' << ppm0.height << " 255" << std::endl;
	int diff = 0;
	int maxerror = 0;
	
	for (unsigned i=0; i<ppm0.pixels.size(); i++) {
		if (ppm0.pixels[i] != ppm1.pixels[i]) {
			diff++;
			maxerror = std::max(max_error(ppm0.pixels[i], ppm1.pixels[i]), maxerror);
		}
		
		std::cout << viewer->diff(ppm0.pixels[i], ppm1.pixels[i]) << std::endl;
	}
	
	std::cerr << "different pixels: "
		<< diff << '/' << ppm0.pixels.size() << std::endl;
	std::cerr << "maximum absolute error: " << maxerror << std::endl;
}

int main(int argc, char* argv[])
{
	diff_viewer* viewer = 0;
	const char* file0 = 0;
	const char* file1 = 0;
	
	switch (argc)
	{
	case 3:
		viewer = new normal_viewer();
		file0 = argv[1];
		file1 = argv[2];
		break;
	case 4:
		file0 = argv[2];
		file1 = argv[3];
		if (std::string(argv[1]) == "-e") {
			viewer = new enhance_viewer();
			break;
		}
		else if (std::string(argv[1]) == "-c") {
			viewer = new classify_viewer();
			break;
		}
		else if (std::string(argv[1]) == "-v") {
			viewer = new verify_viewer();
			break;
		}
	default:
		usage();
		return 1;
	}

	try {
		ppm ppm0 = read_ppm(file0);
		ppm ppm1 = read_ppm(file1);
		
		compare(viewer, ppm0, ppm1);
	}
	catch (std::runtime_error& e) {
		std::cerr << "error: " << e.what() << std::endl;
		delete viewer;
		return 1;
	}
	
	return 0;
}
